//this is a quick and dirty emulator for the Boris Diplomat
//it was created to confirm the ROM dump
//this is not the correct way to do it for MESS, but is useful for me for testing
//it also lets you look at RAM contents and dumps a coverage map showing what bytes in the ROM were read as opcodes and which were read as data
//it's essentially a simple debugger with specialized commands for sending the game key presses
//it breakpoints and displays the LED digits and accepts inputs

// "  -   00" is the game's initial prompt
//to move from D2 to D4: type "d", enter, "2", enter, "d", enter, "4", enter, "enter", enter 
//type board to see the current piece positions

//use help command to list commands available

#include "stdafx.h"
using namespace System;
using namespace System::IO;
using namespace System::Diagnostics;

//#define DEBUGFILE 1 //to dump trace to file f8debug.txt
#define DEBUGFILE 0 //no trace file

//#define DEBUGPC0 1 //to dump pc0 to file f8pc0.txt
#define DEBUGPC0 0 //no pc0 dump

#define ADDRESSMASK 0x7ff	// SKR - 3870 has 11-bit address space, but Boris code uses $5000-57ff

#define _CRT_SECURE_NO_WARNINGS 1
/*****************************************************************************
 *
 *	 f8.c
 *	 Portable F8 emulator (Fairchild 3850)
 *
 *	 Copyright (c) 2000 Juergen Buchmueller, all rights reserved.
 *
 *	 - This source code is released as freeware for non-commercial purposes.
 *	 - You are free to use and redistribute this code in modified or
 *	   unmodified form, provided you list me in the credits.
 *	 - If you modify this source code, you must add a notice to each modified
 *	   source file that it has been changed.  If you're a nice person, you
 *	   will clearly mark each change too.  :)
 *	 - If you wish to use this for commercial purposes, please contact me at
 *	   pullmoll@t-online.de
 *	 - The author of this copywritten work reserves the right to change the
 *     terms of its usage and license at any time, including retroactively
 *   - This entire notice must remain in the source code.
 *
 *	This work is based on Frank Palazzolo's F8 emulation in a standalone
 *	Fairchild Channel F emulator and the 'Fairchild F3850 CPU' data sheets.
 *
 *****************************************************************************/

/* PeT 25.June 2001
   added interrupt functionality
 */

//SKR - lots of mods; a few bug fixes, 3870 tweaks

#include <stdio.h>
#include <stdlib.h>

#define UINT32 Int32
#define UINT16 Int16
#define UINT8 Byte
#define INT8 SByte

#define ROMSIZE 2048
#define RAMSIZE 256
#define NUMPORTS 256
unsigned char ports[NUMPORTS];
unsigned char ram[RAMSIZE];
unsigned char rom[ROMSIZE];
unsigned char access[ROMSIZE];
unsigned int numaccesses[ROMSIZE];
unsigned char octal;
UINT8 ramaddr;
bool singlestep=false;
#define NUMBPS 0
UINT16 bps[NUMBPS+1];
bool show;
UINT8 keycol;
UINT8 keyrow;
UINT8 keydown;
unsigned char digits[8];
unsigned char olddigs[8];

UINT8 cpu_readop(UINT16 addr) {
	addr&=ADDRESSMASK;		//11-bit address space
	access[addr]|=1;	//mark coverage map opcode
	numaccesses[addr]++;
	return rom[addr];
}
UINT8 cpu_readop_arg(UINT16 addr) {
	addr&=ADDRESSMASK;		//11-bit address space
	access[addr]|=2;	//mark coverage map arg
	numaccesses[addr]++;
	return rom[addr];
}
UINT8 program_read_byte_8(UINT16 addr) {
	addr&=ADDRESSMASK;		//11-bit address space
	access[addr]|=4;	//mark coverage map data
	numaccesses[addr]++;
	return rom[addr];
}
void program_write_byte_8(UINT16 addr,UINT8 data) {
	addr&=ADDRESSMASK;		//11-bit address space
	access[addr]|=8;	//mark coverage map write data
	numaccesses[addr]++;
	//can't write to ROM, but mark the coverage map
}


//this converts the LED segment data to ASCII
void showdig(UINT8 segdata) {
	if((segdata&0x7f)==0x77)	//cursor
		Console::Write("_");
	else if((segdata&0x7f)==0x2c)	//?
		Console::Write("?");
	else if((segdata&0x7f)==0x7e)	//-
		Console::Write("-");
	else if((segdata&0x7f)==0x0b)	//h
		Console::Write("h");
	else if((segdata&0x7f)==0x7f)	//space
		Console::Write(" ");
	else if((segdata&0x7f)==0x07)	//t
		Console::Write("t");

//	else if((segdata&0x7f)==0x08)	//black King	-- same symbol as A
//		Console::Write("k");
	else if((segdata&0x7f)==0x48)	//black Queen
		Console::Write("q");
	else if((segdata&0x7f)==0x1c)	//black Rook
		Console::Write("r");
	else if((segdata&0x7f)==0x58)	//black Bishop
		Console::Write("b");
	else if((segdata&0x7f)==0x1a)	//black Knight
		Console::Write("n");
	else if((segdata&0x7f)==0x5c)	//black Pawn
		Console::Write("p");

	else if((segdata&0x7f)==0x63)	//white Pawn
		Console::Write("P");
	else if((segdata&0x7f)==0x13)	//white Knight
		Console::Write("K");
	else if((segdata&0x7f)==0x43)	//white Bishop
		Console::Write("B");
	else if((segdata&0x7f)==0x23)	//white Rook
		Console::Write("R");
	else if((segdata&0x7f)==0x41)	//white Queen
		Console::Write("Q");
	else if((segdata&0x7f)==0x01)	//white King
		Console::Write("K");

	else if((segdata&0x7f)==0x40)	//0
		Console::Write("0");
	else if((segdata&0x7f)==0x79)	//1
		Console::Write("1");
	else if((segdata&0x7f)==0x24)	//2
		Console::Write("2");
	else if((segdata&0x7f)==0x30)	//3
		Console::Write("3");
	else if((segdata&0x7f)==0x19)	//4
		Console::Write("4");
	else if((segdata&0x7f)==0x12)	//5  also S
		Console::Write("5");
	else if((segdata&0x7f)==0x02)	//6
		Console::Write("6");
	else if((segdata&0x7f)==0x78)	//7
		Console::Write("7");
	else if((segdata&0x7f)==0x00)	//8
		Console::Write("8");
	else if((segdata&0x7f)==0x10)	//9
		Console::Write("9");

	else if((segdata&0x7f)==0x08)	//A
		Console::Write("A");
	else if((segdata&0x7f)==0x03)	//b
		Console::Write("b");
	else if((segdata&0x7f)==0x46)	//C
		Console::Write("C");
	else if((segdata&0x7f)==0x21)	//d
		Console::Write("d");
	else if((segdata&0x7f)==0x06)	//E
		Console::Write("E");
	else if((segdata&0x7f)==0x0e)	//F
		Console::Write("F");
	else if((segdata&0x7f)==0x42)	//G
		Console::Write("G");
	else if((segdata&0x7f)==0x09)	//H
		Console::Write("H");
	else
		Console::Write(segdata.ToString("X2"));
	Console::Write(" ");
}


typedef struct {
	UINT16	pc0;	/* program counter 0 */
	UINT16	pc1;	/* program counter 1 */
	UINT16	dc0;	/* data counter 0 */
	UINT16	dc1;	/* data counter 1 */
	UINT8	a;		/* accumulator */
	UINT8	w;		/* processor status */
	UINT8	is; 	/* scratchpad pointer */
	UINT8	dbus;	/* data bus value */
	UINT8	io; 	/* last I/O address */
    UINT16  irq_vector;
	int 	(*irq_callback)(int irqline);
    UINT8   r[64];  /* scratchpad RAM */
    int     irq_request;

	UINT16	oldpc0;	/* prev program counter 0 */

	//SKR - 3870 timer
	UINT8	timercount;
	UINT8	timerrunning;
	UINT8	timerintenabled;
	UINT8	timerdivisor;
	UINT8	timermodulon;
	UINT8	timerirq;
	UINT8	timersubcount;	//when subcount reaches divisor, decrement count

}	f8_Regs;

//port 0=keycols and octal
//port 1=LED segments			
//port 4=RAM data
//port 5=RAM addr

UINT8 io_read_byte_8(UINT8 port,f8_Regs* f8) {
	UINT8 data;

	data=ports[port];	//default return value - whatever was written to port last
	if(port==4) {	//RAM databus
		data=ram[ramaddr];
	} else if(port==0) {	//keycols and octal
		if(octal==keyrow) {
			data=keycol|(ports[port]&0x07);
			if(keydown++==3) {	//stretch out key presses by duplicating data for multiple port reads
				keydown=0;
				keycol=0;
				keyrow=9;
			}
		}
	} else if(port==7)
		data=f8->timercount;
	else if(port==6)
		data=0;		//don't read ICP, return status of ext int pin in bit 0

	if(show)
		Console::WriteLine("io read port {0}, data {1}",port,data);
	return data;
}

void io_write_byte_8(UINT8 port,UINT8 data,f8_Regs *f8) {
	ports[port]=data;
	if(show)
		Console::WriteLine(String::Concat("io write ",port," ",data.ToString("X2")));
	if(port==5) {	//RAM address bus
		ramaddr=data;
		if(show)
			Console::WriteLine(String::Concat("RAM Addr:",data.ToString("X2")));
	} else if(port==4) {	//RAM data bus
		ram[ramaddr]=data;
		if(show)
			Console::WriteLine(String::Concat("RAM Data:",data.ToString("X2")));
	} else if(port==1) {
		digits[octal]=data;
	} else if(port==0) {
		octal=data&0x7;
	} else if(port==7) {//timer count
		f8->timercount=data;
		f8->timermodulon=data;
		f8->timerirq=0;

//		f8->timerdivisor=1;	//docs say "prescaler reset", but I don't think they mean "divisor set to 1"
		f8->timersubcount=0;
	} else if(port==6) {//timer ICP
		f8->timerdivisor=1;
		f8->timerintenabled=((data&0x02)==0x02);
		f8->timerrunning=((data&0x08)==0x08);
		if(f8->timerrunning==0)
			f8->timersubcount=0;
		if(data&0x20)
			f8->timerdivisor*=2;
		if(data&0x40)
			f8->timerdivisor*=5;
		if(data&0x80)
			f8->timerdivisor*=20;
	}
}

enum
{
	F8_PC0=1, F8_PC1, F8_DC0, F8_DC1, F8_W, F8_A, F8_IS,
        F8_J, F8_HU, F8_HL, F8_KU, F8_KL, F8_QU, F8_QL,

        F8_R0, F8_R1, F8_R2, F8_R3, F8_R4, F8_R5, F8_R6, F8_R7, F8_R8,
        F8_R16, F8_R17, F8_R18, F8_R19, F8_R20, F8_R21, F8_R22, F8_R23,
        F8_R24, F8_R25, F8_R26, F8_R27, F8_R28, F8_R29, F8_R30, F8_R31,
        F8_R32, F8_R33, F8_R34, F8_R35, F8_R36, F8_R37, F8_R38, F8_R39,
        F8_R40, F8_R41, F8_R42, F8_R43, F8_R44, F8_R45, F8_R46, F8_R47,
        F8_R48, F8_R49, F8_R50, F8_R51, F8_R52, F8_R53, F8_R54, F8_R55,
        F8_R56, F8_R57, F8_R58, F8_R59, F8_R60, F8_R61, F8_R62, F8_R63
};

static const char *const rname[16] = {
	"R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7",
	"R8", "J",  "HU", "HL", "KU", "KL", "QU", "QL"
};

int f8dasm(unsigned char *rom, int pc,char *buffer) {
	unsigned size = 0;
	UINT8 op = rom[(pc+size++)&ADDRESSMASK];

    switch( op ) {
	/* opcode  bitmask */
	case 0x00: /* 0000 0000 */
        sprintf(buffer, "LR   A,KU");
		break;
	case 0x01: /* 0000 0001 */
        sprintf(buffer, "LR   A,KL");
		break;
	case 0x02: /* 0000 0010 */
        sprintf(buffer, "LR   A,QU");
		break;
	case 0x03: /* 0000 0011 */
        sprintf(buffer, "LR   A,QL");
		break;
	case 0x04: /* 0000 0100 */
        sprintf(buffer, "LR   KU,A");
		break;
	case 0x05: /* 0000 0101 */
        sprintf(buffer, "LR   KL,A");
		break;
	case 0x06: /* 0000 0110 */
        sprintf(buffer, "LR   QU,A");
		break;
	case 0x07: /* 0000 0111 */
        sprintf(buffer, "LR   QL,A");
		break;

    case 0x08: /* 0000 1000 */
        sprintf(buffer, "LR   K,P");
		break;
	case 0x09: /* 0000 1001 */
        sprintf(buffer, "LR   P,K");
		break;
	case 0x0a: /* 0000 1010 */
		sprintf(buffer, "LR   A,IS");
		break;
	case 0x0b: /* 0000 1011 */
		sprintf(buffer, "LR   IS,A");
		break;

    case 0x0c: /* 0000 1100 */
		sprintf(buffer, "PK") ;
		break;
	case 0x0d: /* 0000 1101 */
		sprintf(buffer, "LR   P0,Q");
        break;
	case 0x0e: /* 0000 1110 */
        sprintf(buffer, "LR   Q,DC");
		break;
	case 0x0f: /* 0000 1111 */
        sprintf(buffer, "LR   DC,Q");
		break;
	case 0x10: /* 0001 0000 */
		sprintf(buffer, "LR   DC,H");
		break;
	case 0x11: /* 0001 0001 */
		sprintf(buffer, "LR   H,DC");
		break;
	case 0x12: /* 0001 0010 */
		sprintf(buffer, "SR   1");
		break;
	case 0x13: /* 0001 0011 */
		sprintf(buffer, "SL   1");
		break;
	case 0x14: /* 0001 0100 */
		sprintf(buffer, "SR   4");
		break;
	case 0x15: /* 0001 0101 */
		sprintf(buffer, "SL   4");
		break;
	case 0x16: /* 0001 0110 */
		sprintf(buffer, "LM") ;
		break;
	case 0x17: /* 0001 0111 */
		sprintf(buffer, "ST");
		break;
	case 0x18: /* 0001 1000 */
		sprintf(buffer, "COM");
		break;
	case 0x19: /* 0001 1001 */
		sprintf(buffer, "LNK");
		break;
	case 0x1a: /* 0001 1010 */
		sprintf(buffer, "DI");
		break;
	case 0x1b: /* 0001 1011 */
		sprintf(buffer, "EI");
		break;
	case 0x1c: /* 0001 1100 */
        sprintf(buffer, "POP");
		break;
	case 0x1d: /* 0001 1101 */
        sprintf(buffer, "LR   W,J");
		break;
	case 0x1e: /* 0001 1110 */
        sprintf(buffer, "LR   J,W");
		break;
	case 0x1f: /* 0001 1111 */
		sprintf(buffer, "INC");
		break;
	case 0x20: /* 0010 0000 */
		sprintf(buffer, "LI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
		break;
	case 0x21: /* 0010 0001 */
		sprintf(buffer, "NI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
    case 0x22: /* 0010 0010 */
		sprintf(buffer, "OI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
    case 0x23: /* 0010 0011 */
		sprintf(buffer, "XI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
	case 0x24: /* 0010 0100 */
		sprintf(buffer, "AI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
	case 0x25: /* 0010 0101 */
		sprintf(buffer, "CI   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
	case 0x26: /* 0010 0110 */
		sprintf(buffer, "IN   $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
	case 0x27: /* 0010 0111 */
		sprintf(buffer, "OUT  $%02X", rom[(pc+size++)&ADDRESSMASK]);
        break;
	case 0x28: /* 0010 1000 */
		sprintf(buffer, "PI   $%02X%02X", rom[(pc+size + 0)&ADDRESSMASK], rom[(pc+size + 1)&ADDRESSMASK]);
		size += 2;
        break;
	case 0x29: /* 0010 1001 */
		sprintf(buffer, "JMP  $%02X%02X", rom[(pc+size + 0)&ADDRESSMASK], rom[(pc+size + 1)&ADDRESSMASK]);
		size += 2;
        break;
	case 0x2a: /* 0010 1010 */
		sprintf(buffer, "DCI  $%02X%02X", rom[(pc+size + 0)&ADDRESSMASK], rom[(pc+size + 1)&ADDRESSMASK]);
		size += 2;
        break;
	case 0x2b: /* 0010 1011 */
		sprintf(buffer, "NOP");
		break;
	case 0x2c: /* 0010 1100 */
		sprintf(buffer, "XDC");
		break;
	case 0x2d: /* 0010 1101 */
	case 0x2e: /* 0010 1110 */
	case 0x2f: /* 0010 1111 */
		sprintf(buffer, "???  $%02X",op);
        break;

    case 0x30: /* 0011 0000 */
	case 0x31: /* 0011 0001 */
	case 0x32: /* 0011 0010 */
	case 0x33: /* 0011 0011 */
	case 0x34: /* 0011 0100 */
	case 0x35: /* 0011 0101 */
	case 0x36: /* 0011 0110 */
	case 0x37: /* 0011 0111 */
	case 0x38: /* 0011 1000 */
	case 0x39: /* 0011 1001 */
	case 0x3a: /* 0011 1010 */
	case 0x3b: /* 0011 1011 */
        sprintf(buffer, "DS   %s",rname[op & 15]);
        break;
    case 0x3c: /* 0011 1100 */
        sprintf(buffer, "DS   (IS)");
        break;
    case 0x3d: /* 0011 1101 */
        sprintf(buffer, "DS   (IS++)");
        break;
    case 0x3e: /* 0011 1110 */
        sprintf(buffer, "DS   (IS--)");
        break;
    case 0x3f: /* 0011 1111 */
		sprintf(buffer, "???  $%02X",op);
        break;

    case 0x40: /* 0100 0000 */
    case 0x41: /* 0100 0001 */
	case 0x42: /* 0100 0010 */
	case 0x43: /* 0100 0011 */
	case 0x44: /* 0100 0100 */
	case 0x45: /* 0100 0101 */
	case 0x46: /* 0100 0110 */
	case 0x47: /* 0100 0111 */
	case 0x48: /* 0100 1000 */
	case 0x49: /* 0100 1001 */
	case 0x4a: /* 0100 1010 */
	case 0x4b: /* 0100 1011 */
        sprintf(buffer, "LR   A,%s",rname[op & 15]);
        break;
    case 0x4c: /* 0100 1100 */
        sprintf(buffer, "LR   A,(IS)");
        break;
    case 0x4d: /* 0100 1101 */
        sprintf(buffer, "LR   A,(IS++)");
        break;
    case 0x4e: /* 0100 1110 */
        sprintf(buffer, "LR   A,(IS--)");
        break;
    case 0x4f: /* 0100 1111 */
		sprintf(buffer, "???  $%02X",op);
        break;

    case 0x50: /* 0101 0000 */
	case 0x51: /* 0101 0001 */
	case 0x52: /* 0101 0010 */
	case 0x53: /* 0101 0011 */
	case 0x54: /* 0101 0100 */
	case 0x55: /* 0101 0101 */
	case 0x56: /* 0101 0110 */
	case 0x57: /* 0101 0111 */
	case 0x58: /* 0101 1000 */
	case 0x59: /* 0101 1001 */
	case 0x5a: /* 0101 1010 */
	case 0x5b: /* 0101 1011 */
        sprintf(buffer, "LR   %s,A",rname[op & 15]);
        break;
    case 0x5c: /* 0101 1100 */
        sprintf(buffer, "LR   (IS),A");
        break;
    case 0x5d: /* 0101 1101 */
        sprintf(buffer, "LR   (IS++),A");
        break;
    case 0x5e: /* 0101 1110 */
        sprintf(buffer, "LR   (IS--),A");
        break;
    case 0x5f: /* 0101 1111 */
		sprintf(buffer, "???  $%02X",op);
        break;

    case 0x60: /* 0110 0000 */
    case 0x61: /* 0110 0001 */
	case 0x62: /* 0110 0010 */
	case 0x63: /* 0110 0011 */
	case 0x64: /* 0110 0100 */
	case 0x65: /* 0110 0101 */
	case 0x66: /* 0110 0110 */
	case 0x67: /* 0110 0111 */
		sprintf(buffer, "LISU $%02X", op & 0x07);
		break;
    case 0x68: /* 0110 1000 */
	case 0x69: /* 0110 1001 */
	case 0x6a: /* 0110 1010 */
	case 0x6b: /* 0110 1011 */
	case 0x6c: /* 0110 1100 */
	case 0x6d: /* 0110 1101 */
	case 0x6e: /* 0110 1110 */
	case 0x6f: /* 0110 1111 */
		sprintf(buffer, "LISL $%02X", op & 0x07);
        break;

    case 0x70: /* 0111 0000 */
	case 0x71: /* 0111 0001 */
	case 0x72: /* 0111 0010 */
	case 0x73: /* 0111 0011 */
	case 0x74: /* 0111 0100 */
	case 0x75: /* 0111 0101 */
	case 0x76: /* 0111 0110 */
	case 0x77: /* 0111 0111 */
	case 0x78: /* 0111 1000 */
	case 0x79: /* 0111 1001 */
	case 0x7a: /* 0111 1010 */
	case 0x7b: /* 0111 1011 */
	case 0x7c: /* 0111 1100 */
	case 0x7d: /* 0111 1101 */
	case 0x7e: /* 0111 1110 */
	case 0x7f: /* 0111 1111 */
		sprintf(buffer, "LIS  $%02X", op & 0x0f);
		break;

	case 0x81: /* 1000 0001 */
	case 0x85: /* 1000 0101 */
		sprintf(buffer, "BP   $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x82: /* 1000 0010 */
		sprintf(buffer, "BC   $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x84: /* 1000 0100 */
		sprintf(buffer, "BZ   $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x80: /* 1000 0000 */
	case 0x83: /* 1000 0011 */
	case 0x86: /* 1000 0110 */
	case 0x87: /* 1000 0111 */
		sprintf(buffer, "BT   $%02X,$%04X", op & 0x07, pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

    case 0x88: /* 1000 1000 */
		sprintf(buffer, "AM");
		break;

    case 0x89: /* 1000 1001 */
		sprintf(buffer, "AMD");
        break;

    case 0x8a: /* 1000 1010 */
		sprintf(buffer, "NM");
		break;

    case 0x8b: /* 1000 1011 */
		sprintf(buffer, "OM");
		break;

    case 0x8c: /* 1000 1100 */
		sprintf(buffer, "XM");
		break;

    case 0x8d: /* 1000 1101 */
		sprintf(buffer, "CM");
		break;

    case 0x8e: /* 1000 1110 */
		sprintf(buffer, "ADC");
        break;

    case 0x8f: /* 1000 1111 */
		sprintf(buffer, "BR7  $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

    case 0x90: /* 1001 0000 */
		sprintf(buffer, "BR   $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x91: /* 1001 0001 */
	case 0x95: /* 1001 0101 */
		sprintf(buffer, "BM   $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x92: /* 1001 0010 */
		sprintf(buffer, "BNC  $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x94: /* 1001 0100 */
		sprintf(buffer, "BNZ  $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x98: /* 1001 1000 */
		sprintf(buffer, "BNO  $%04X", pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

	case 0x93: /* 1001 0011 */
	case 0x96: /* 1001 0110 */
	case 0x97: /* 1001 0111 */
	case 0x99: /* 1001 1001 */
	case 0x9a: /* 1001 1010 */
	case 0x9b: /* 1001 1011 */
	case 0x9c: /* 1001 1100 */
	case 0x9d: /* 1001 1101 */
	case 0x9e: /* 1001 1110 */
	case 0x9f: /* 1001 1111 */
		sprintf(buffer, "BF   $%02X,$%04X", op & 0x0f, pc + (INT8)rom[(pc+size++)&ADDRESSMASK] + 1);
        break;

    case 0xa0: /* 1010 0000 */
	case 0xa1: /* 1010 0001 */
		sprintf(buffer, "INS  $%02X", (unsigned) (INT8) (op & 0x0F));
        break;

    case 0xa2: /* 1010 0010 */
	case 0xa3: /* 1010 0011 */
		sprintf(buffer, "???  $%02X\n", op);
        break;

    case 0xa4: /* 1010 0100 */
	case 0xa5: /* 1010 0101 */
	case 0xa6: /* 1010 0110 */
	case 0xa7: /* 1010 0111 */
	case 0xa8: /* 1010 1000 */
	case 0xa9: /* 1010 1001 */
	case 0xaa: /* 1010 1010 */
	case 0xab: /* 1010 1011 */
	case 0xac: /* 1010 1100 */
	case 0xad: /* 1010 1101 */
	case 0xae: /* 1010 1110 */
	case 0xaf: /* 1010 1111 */
		sprintf(buffer, "INS  $%02X", (INT8) op & 0x0f);
        break;

    case 0xb0: /* 1011 0000 */
	case 0xb1: /* 1011 0001 */
		sprintf(buffer, "OUTS $%02X", (INT8) op & 0x0f);
        break;

    case 0xb2: /* 1011 0010 */
	case 0xb3: /* 1011 0011 */
		sprintf(buffer, "???  $%02X\n", op);
        break;

    case 0xb4: /* 1011 0100 */
	case 0xb5: /* 1011 0101 */
	case 0xb6: /* 1011 0110 */
	case 0xb7: /* 1011 0111 */
	case 0xb8: /* 1011 1000 */
	case 0xb9: /* 1011 1001 */
	case 0xba: /* 1011 1010 */
	case 0xbb: /* 1011 1011 */
	case 0xbc: /* 1011 1100 */
	case 0xbd: /* 1011 1101 */
	case 0xbe: /* 1011 1110 */
	case 0xbf: /* 1011 1111 */
		sprintf(buffer, "OUTS $%02X", (unsigned) (INT8) op & 0x0f);
        break;

    case 0xc0: /* 1100 0000 */
	case 0xc1: /* 1100 0001 */
	case 0xc2: /* 1100 0010 */
	case 0xc3: /* 1100 0011 */
	case 0xc4: /* 1100 0100 */
	case 0xc5: /* 1100 0101 */
	case 0xc6: /* 1100 0110 */
	case 0xc7: /* 1100 0111 */
	case 0xc8: /* 1100 1000 */
	case 0xc9: /* 1100 1001 */
	case 0xca: /* 1100 1010 */
	case 0xcb: /* 1100 1011 */
        sprintf(buffer, "AS   %s", rname[op & 15]);
        break;
    case 0xcc: /* 1100 1100 */
		sprintf(buffer, "AS   (IS)");
        break;
    case 0xcd: /* 1100 1101 */
		sprintf(buffer, "AS   (IS++)");
        break;
    case 0xce: /* 1100 1110 */
		sprintf(buffer, "AS   (IS--)");
        break;
    case 0xcf: /* 1100 1111 */
		sprintf(buffer, "???  $%02X\n", op);
		break;

    case 0xd0: /* 1101 0000 */
	case 0xd1: /* 1101 0001 */
	case 0xd2: /* 1101 0010 */
	case 0xd3: /* 1101 0011 */
	case 0xd4: /* 1101 0100 */
	case 0xd5: /* 1101 0101 */
	case 0xd6: /* 1101 0110 */
	case 0xd7: /* 1101 0111 */
	case 0xd8: /* 1101 1000 */
	case 0xd9: /* 1101 1001 */
	case 0xda: /* 1101 1010 */
	case 0xdb: /* 1101 1011 */
		sprintf(buffer, "ASD  %s", rname[op & 15]);
		break;
    case 0xdc: /* 1101 1100 */
		sprintf(buffer, "ASD  (IS)");
		break;
    case 0xdd: /* 1101 1101 */
		sprintf(buffer, "ASD  (IS++)");
        break;
    case 0xde: /* 1101 1110 */
		sprintf(buffer, "ASD  (IS--)");
        break;
    case 0xdf: /* 1101 1111 */
		sprintf(buffer, "???  $%02X\n", op);
        break;

    case 0xe0: /* 1110 0000 */
	case 0xe1: /* 1110 0001 */
	case 0xe2: /* 1110 0010 */
	case 0xe3: /* 1110 0011 */
	case 0xe4: /* 1110 0100 */
	case 0xe5: /* 1110 0101 */
	case 0xe6: /* 1110 0110 */
	case 0xe7: /* 1110 0111 */
	case 0xe8: /* 1110 1000 */
	case 0xe9: /* 1110 1001 */
	case 0xea: /* 1110 1010 */
	case 0xeb: /* 1110 1011 */
		sprintf(buffer, "XS   %s", rname[op & 15]);
        break;
    case 0xec: /* 1110 1100 */
		sprintf(buffer, "XS   (IS)");
        break;
    case 0xed: /* 1110 1101 */
		sprintf(buffer, "XS   (IS++)");
        break;
    case 0xee: /* 1110 1110 */
		sprintf(buffer, "XS   (IS--)");
        break;
    case 0xef: /* 1110 1111 */
		sprintf(buffer, "???  $%02X\n", op);
        break;


    case 0xf0: /* 1111 0000 */
	case 0xf1: /* 1111 0001 */
	case 0xf2: /* 1111 0010 */
	case 0xf3: /* 1111 0011 */
	case 0xf4: /* 1111 0100 */
	case 0xf5: /* 1111 0101 */
	case 0xf6: /* 1111 0110 */
	case 0xf7: /* 1111 0111 */
	case 0xf8: /* 1111 1000 */
	case 0xf9: /* 1111 1001 */
	case 0xfa: /* 1111 1010 */
	case 0xfb: /* 1111 1011 */
		sprintf(buffer, "NS   %s", rname[op & 15]);
		break;
    case 0xfc: /* 1111 1100 */
		sprintf(buffer, "NS   (IS)");
        break;
    case 0xfd: /* 1111 1101 */
		sprintf(buffer, "NS   (IS++)");
        break;
    case 0xfe: /* 1111 1110 */
		sprintf(buffer, "NS   (IS--)");
        break;
    case 0xff: /* 1111 1111 */
		sprintf(buffer, "???  $%02X\n", op);
        break;
    }//switch(op)

    return size;
}//f8dasm

//status register bits
#define W_S	0x01	//sign
#define W_C	0x02	//carry
#define W_Z	0x04	//zero
#define W_O	0x08	//overflow
#define W_I	0x10	//icb

//instruction timing
#define cS  4		//short
#define cL	6		//long

static f8_Regs f8;

//status register macros
/* clear all flags */
#define CLR_OZCS                \
	f8.w &= ~(W_O|W_Z|W_C|W_S)

/* set sign and zero flags (note: the S flag is complementary) */
#define SET_SZ(n)               \
	if (n == 0) 				\
		f8.w |= W_Z | W_S;			\
	else						\
	if (n < 128)				\
		f8.w |= W_S

/* set overflow and carry flags */
#define SET_OC(n,m)             \
	if (n + m > 255)			\
		f8.w |= W_C;				\
	if ((n&127)+(m&127) > 127)	\
	{							\
		if (!(f8.w & W_C))		\
			f8.w |= W_O;			\
	}							\
	else						\
	{							\
		if (f8.w & W_C)			\
			f8.w |= W_O;			\
	}


void updatetimer(UINT8 cycles) {	//after each ROMC, update timer
	if(f8.timerrunning) {							//is timer running?
		f8.timersubcount+=cycles;					//
		while(f8.timersubcount>=f8.timerdivisor) {	//prescaler
			f8.timersubcount-=f8.timerdivisor;
			f8.timercount--;
			if(f8.timercount==0) {
				f8.timercount=f8.timermodulon;
				f8.timerirq=1;
			}
		}
	}
}

/******************************************************************************
 * ROMC (ROM cycles)
 * This is what the Fairchild F8 CPUs use instead of an address bus
 * There are 5 control lines and each combination of those lines has
 * a special meaning. The devices attached to those control lines all
 * have their own program counters (PC0 and PC1) and at least one
 * data counter (DC0).
 * Currently the emulation does not handle distinct PCs and DCs, but
 * only one instance inside the CPU context.
 ******************************************************************************/
static void ROMC_00(int insttim)	/* SKR - added parameter to tell if  */
                                    /* it is long or short based on inst */
{
    /*
     * Instruction Fetch. The device whose address space includes the
     * contents of the PC0 register must place on the data bus the op
     * code addressed by PC0; then all devices increment the contents
     * of PC0.
     */

	f8.dbus = cpu_readop(f8.pc0);
    f8.pc0 += 1;
	/* SKR - ROMC00 is usually short, not short+long, but DS is long */
	updatetimer(insttim);
}

static void ROMC_01(void)
{
    /*
     * The device whose address space includes the contents of the PC0
     * register must place on the data bus the contents of the memory
     * location addressed by PC0; then all devices add the 8-bit value
     * on the data bus as signed binary number to PC0.
     */
	f8.dbus = cpu_readop_arg(f8.pc0);
	f8.pc0 += (INT8)f8.dbus;
	updatetimer(cL);
}

static void ROMC_02(void)
{
    /*
     * The device whose DC0 addresses a memory word within the address
     * space of that device must place on the data bus the contents of
     * the memory location addressed by DC0; then all devices increment
     * DC0.
     */
    f8.dbus = program_read_byte_8(f8.dc0);
    f8.dc0 += 1;
	updatetimer(cL);
}

static void ROMC_03(int insttim)	/* SKR - added parameter to tell if  */
{                                   /* it is long or short based on inst */
    /*
     * Similiar to 0x00, except that it is used for immediate operands
     * fetches (using PC0) instead of instruction fetches.
     */
    f8.dbus = f8.io = cpu_readop_arg(f8.pc0);
    f8.pc0 += 1;
	updatetimer(insttim);
}

static void ROMC_04(void)
{
    /*
     * Copy the contents of PC1 into PC0
     */
    f8.pc0 = f8.pc1;
	updatetimer(cS);
}

static void ROMC_05(void)
{
    /*
     * Store the data bus contents into the memory location pointed
     * to by DC0; increment DC0.
     */
    program_write_byte_8(f8.dc0, f8.dbus);
    f8.dc0 += 1;
	updatetimer(cL);
}

static void ROMC_06(void)
{
    /*
     * Place the high order byte of DC0 on the data bus.
     */
    f8.dbus = f8.dc0 >> 8;
	updatetimer(cL);
}

static void ROMC_07(void)
{
    /*
     * Place the high order byte of PC1 on the data bus.
     */
    f8.dbus = f8.pc1 >> 8;
	updatetimer(cL);
}

static void ROMC_08(void)
{
    /*
     * All devices copy the contents of PC0 into PC1. The CPU outputs
     * zero on the data bus in this ROMC state. Load the data bus into
     * both halves of PC0, thus clearing the register.
     */
    f8.pc1 = f8.pc0;
    f8.dbus = 0;
    f8.pc0 = 0;
	updatetimer(cL);
}

static void ROMC_09(void)
{
    /*
     * The device whose address space includes the contents of the DC0
     * register must place the low order byte of DC0 onto the data bus.
     */
    f8.dbus = f8.dc0 & 0xff;
	updatetimer(cL);
}

static void ROMC_0A(void)
{
    /*
     * All devices add the 8-bit value on the data bus, treated as
     * signed binary number, to the data counter.
     */
	f8.dc0 += (INT8)f8.dbus;
	updatetimer(cL);
}

static void ROMC_0B(void)
{
    /*
     * The device whose address space includes the value in PC1
     * must place the low order byte of PC1 onto the data bus.
     */
    f8.dbus = f8.pc1 & 0xff;
	updatetimer(cL);
}

static void ROMC_0C(void)
{
    /*
     * The device whose address space includes the contents of the PC0
     * register must place the contents of the memory word addressed
     * by PC0 into the data bus; then all devices move the value that
     * has just been placed on the data bus into the low order byte of PC0.
     */
    f8.dbus = cpu_readop_arg(f8.pc0);
    f8.pc0 = (f8.pc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_0D(void)
{
    /*
     * All devices store in PC1 the current contents of PC0, incremented
     * by 1; PC0 is unaltered.
     */
    f8.pc1 = f8.pc0 + 1;
	updatetimer(cS);
}

static void ROMC_0E(void)
{
    /*
     * The device whose address space includes the contents of the PC0
     * register must place the word addressed by PC0 into the data bus.
     * The value on the data bus is then moved to the low order byte
     * of DC0 by all devices.
     */
    f8.dbus = cpu_readop_arg(f8.pc0);
    f8.dc0 = (f8.dc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_0F(void)
{
    /*
     * The interrupting device with highest priority must place the
     * low order byte of the interrupt vector on the data bus.
     * All devices must copy the contents of PC0 into PC1. All devices
     * must move the contents of the data bus into the low order
     * byte of PC0.
     */
//    f8.irq_vector = (*f8.irq_callback)(0);
    f8.dbus = f8.irq_vector & 0x00ff;
    f8.pc1 = f8.pc0;
    f8.pc0 = (f8.pc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

#if 0
static void ROMC_10(void)
{
    /*
     * Inhibit any modification to the interrupt priority logic.
     */
    f8.w |= 0x20;   /* ???? */
	updatetimer(cL);
}
#endif

static void ROMC_11(void)
{
    /*
     * The device whose address space includes the contents of PC0
     * must place the contents of the addressed memory word on the
     * data bus. All devices must then move the contents of the
     * data bus to the upper byte of DC0.
     */
    f8.dbus = cpu_readop_arg(f8.pc0);
	f8.dc0 = (f8.dc0 & 0x00ff) | (f8.dbus << 8);
	updatetimer(cL);
}

static void ROMC_12(void)
{
    /*
     * All devices copy the contents of PC0 into PC1. All devices then
     * move the contents of the data bus into the low order byte of PC0.
     */
    f8.pc1 = f8.pc0;
    f8.pc0 = (f8.pc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_13(void)
{
    /*
     * The interrupting device with highest priority must move the high
     * order half of the interrupt vector onto the data bus. All devices
     * must then move the contents of the data bus into the high order
     * byte of PC0. The interrupting device resets its interrupt circuitry
     * (so that it is no longer requesting CPU servicing and can respond
     * to another interrupt).
     */
    f8.dbus = f8.irq_vector >> 8;
    f8.pc0 = (f8.pc0 & 0x00ff) | (f8.dbus << 8);
    f8.w&=~W_I;	//clear ICB
	f8.irq_request=0;
	f8.timerirq=0;
	updatetimer(cL);
}

static void ROMC_14(void)
{
    /*
     * All devices move the contents of the data bus into the high
     * order byte of PC0.
     */
    f8.pc0 = (f8.pc0 & 0x00ff) | (f8.dbus << 8);
	updatetimer(cL);
}

static void ROMC_15(void)
{
    /*
     * All devices move the contents of the data bus into the high
     * order byte of PC1.
     */
    f8.pc1 = (f8.pc1 & 0x00ff) | (f8.dbus << 8);
	updatetimer(cL);
}

static void ROMC_16(void)
{
    /*
     * All devices move the contents of the data bus into the high
     * order byte of DC0.
     */
    f8.dc0 = (f8.dc0 & 0x00ff) | (f8.dbus << 8);
	updatetimer(cL);
}

static void ROMC_17(void)
{
    /*
     * All devices move the contents of the data bus into the low
     * order byte of PC0.
     */
    f8.pc0 = (f8.pc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_18(void)
{
    /*
     * All devices move the contents of the data bus into the low
     * order byte of PC1.
     */
    f8.pc1 = (f8.pc1 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_19(void)
{
    /*
     * All devices move the contents of the data bus into the low
     * order byte of DC0.
     */
    f8.dc0 = (f8.dc0 & 0xff00) | f8.dbus;
	updatetimer(cL);
}

static void ROMC_1A(void)
{
    /*
     * During the prior cycle, an I/O port timer or interrupt control
     * register was addressed; the device containing the addressed port
     * must place the contents of the data bus into the address port.
     */
    io_write_byte_8(f8.io, f8.dbus,&f8);
	updatetimer(cL);
}

static void ROMC_1B(void)
{
    /*
     * During the prior cycle, the data bus specified the address of an
     * I/O port. The device containing the addressed I/O port must place
     * the contents of the I/O port on the data bus. (Note that the
     * contents of timer and interrupt control registers cannot be read
     * back onto the data bus).
     */
	f8.dbus = io_read_byte_8(f8.io,&f8);
	updatetimer(cL);
}

static void ROMC_1C(int insttim)	/* SKR - added parameter to tell if  */
                                    /* it is long or short based on inst */
{
    /*
     * None.
     */
	updatetimer(insttim);
}

static void ROMC_1D(void)
{
    /*
     * Devices with DC0 and DC1 registers must switch registers.
     * Devices without a DC1 register perform no operation.
     */
    UINT16 tmp = f8.dc0;
    f8.dc0 = f8.dc1;
    f8.dc1 = tmp;
	updatetimer(cS);
}

#if 0
static void ROMC_1E(void)
{
    /*
     * The devices whose address space includes the contents of PC0
     * must place the low order byte of PC0 onto the data bus.
     */
    f8.dbus = f8.pc0 & 0xff;
	updatetimer(cL);
}

static void ROMC_1F(void)
{
    /*
     * The devices whose address space includes the contents of PC0
     * must place the high order byte of PC0 onto the data bus.
     */
    f8.dbus = (f8.pc0 >> 8) & 0xff;
	updatetimer(cL);
}
#endif

/***********************************
 *	illegal opcodes
 ***********************************/
static void illegal(void)
{
    Console::WriteLine(String::Concat("f8 illegal opcode at ", f8.pc0.ToString("X4"), f8.dbus.ToString("X2")));
}

/***************************************************
 *	O Z C S 0000 0000
 *	- - - - LR	A,KU
 ***************************************************/
static void f8_lr_a_ku(void)
{
    f8.a = f8.r[12];
}

/***************************************************
 *	O Z C S 0000 0001
 *	- - - - LR	A,KL
 ***************************************************/
static void f8_lr_a_kl(void)
{
    f8.a = f8.r[13];
}

/***************************************************
 *	O Z C S 0000 0010
 *	- - - - LR	A,QU
 ***************************************************/
static void f8_lr_a_qu(void)
{
    f8.a = f8.r[14];
}

/***************************************************
 *	O Z C S 0000 0011
 *	- - - - LR	A,QL
 ***************************************************/
static void f8_lr_a_ql(void)
{
    f8.a = f8.r[15];
}

/***************************************************
 *	O Z C S 0000 0100
 *	- - - - LR	KU,A
 ***************************************************/
static void f8_lr_ku_a(void)
{
    f8.r[12] = f8.a;
}

/***************************************************
 *	O Z C S 0000 0101
 *	- - - - LR	KL,A
 ***************************************************/
static void f8_lr_kl_a(void)
{
    f8.r[13] = f8.a;
}

/***************************************************
 *	O Z C S 0000 0110
 *	- - - - LR	QU,A
 ***************************************************/
static void f8_lr_qu_a(void)
{
    f8.r[14] = f8.a;
}

/***************************************************
 *	O Z C S 0000 0111
 *	- - - - LR	QL,A
 ***************************************************/
static void f8_lr_ql_a(void)
{
    f8.r[15] = f8.a;
}

/***************************************************
 *	O Z C S 0000 1000
 *	- - - - LR	K,P
 ***************************************************/
static void f8_lr_k_p(void)
{
    ROMC_07();
    f8.r[12] = f8.dbus;
    ROMC_0B();
    f8.r[13] = f8.dbus;
}

/***************************************************
 *	O Z C S 0000 1001
 *	- - - - LR	P,K
 ***************************************************/
static void f8_lr_p_k(void)
{
    f8.dbus = f8.r[12];
    ROMC_15();
    f8.dbus = f8.r[13];
    ROMC_18();
}

/***************************************************
 *	O Z C S 0000 1010
 *	- - - - LR	A,IS
 ***************************************************/
static void f8_lr_a_is(void)
{
    f8.a = f8.is;
}

/***************************************************
 *	O Z C S 0000 1011
 *	- - - - LR	IS,A
 ***************************************************/
static void f8_lr_is_a(void)
{
    f8.is = f8.a & 0x3f;
}

/***************************************************
 *	O Z C S 0000 1100
 *	- - - - PK
 ***************************************************/
static void f8_pk(void)
{
    f8.dbus = f8.r[13];
    ROMC_12();
    f8.dbus = f8.r[12];
    ROMC_14();
}

/***************************************************
 *	O Z C S 0000 1101
 *	- - - - LR	P0,Q
 ***************************************************/
static void f8_lr_p0_q(void)
{
    f8.dbus = f8.r[15];
    ROMC_17();
    f8.dbus = f8.r[14];
    ROMC_14();
}

/***************************************************
 *	O Z C S 0000 1110
 *	- - - - LR	 Q,DC
 ***************************************************/
static void f8_lr_q_dc(void)
{
    ROMC_06();
    f8.r[14] = f8.dbus;
    ROMC_09();
    f8.r[15] = f8.dbus;
}

/***************************************************
 *	O Z C S 0000 1111
 *	- - - - LR	 DC,Q
 ***************************************************/
static void f8_lr_dc_q(void)
{
    f8.dbus = f8.r[14];
    ROMC_16();
    f8.dbus = f8.r[15];
    ROMC_19();
}

/***************************************************
 *	O Z C S 0001 0000
 *	- - - - LR	 DC,H
 ***************************************************/
static void f8_lr_dc_h(void)
{
    f8.dbus = f8.r[10];
    ROMC_16();
    f8.dbus = f8.r[11];
    ROMC_19();
}

/***************************************************
 *	O Z C S 0001 0001
 *	- - - - LR	 H,DC
 ***************************************************/
static void f8_lr_h_dc(void)
{
    ROMC_06();
    f8.r[10] = f8.dbus;
    ROMC_09();
    f8.r[11] = f8.dbus;
}

/***************************************************
 *	O Z C S 0001 0010
 *	0 x 0 1 SR	 1
 ***************************************************/
static void f8_sr_1(void)
{
    f8.a >>= 1;
    CLR_OZCS;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 0011
 *	0 x 0 x SL	 1
 ***************************************************/
static void f8_sl_1(void)
{
    f8.a <<= 1;
    CLR_OZCS;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 0100
 *	0 x 0 1 SR	 4
 ***************************************************/
static void f8_sr_4(void)
{
    f8.a >>= 4;
    CLR_OZCS;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 0101
 *	0 x 0 x SL	 4
 ***************************************************/
static void f8_sl_4(void)
{
    f8.a <<= 4;
    CLR_OZCS;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 0110
 *	- - - - LM
 ***************************************************/
static void f8_lm(void)
{
    ROMC_02();
    f8.a = f8.dbus;
}

/***************************************************
 *	O Z C S 0001 0111
 *	- - - - ST
 ***************************************************/
static void f8_st(void)
{
    f8.dbus = f8.a;
    ROMC_05();
}

/***************************************************
 *	O Z C S 0001 1000
 *	0 x 0 x COM
 ***************************************************/
static void f8_com(void)
{
    f8.a = ~f8.a;
    CLR_OZCS;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 1001
 *	x x x x LNK
 ***************************************************/
static void f8_lnk(void)
{
    if (f8.w & W_C)
    {
	CLR_OZCS;
	SET_OC(f8.a,1);
	f8.a += 1;
    }
    else
    {
	CLR_OZCS;
    }
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0001 1010
 *			DI
 ***************************************************/
static void f8_di(void)
{
    ROMC_1C(cS);
    f8.w &= ~W_I;	//clear ICB
}

/***************************************************
 *	O Z C S 0001 1011
 *			EI
 ***************************************************/
static void f8_ei(void)
{
    ROMC_1C(cS);
    f8.w |= W_I;//set ICB
}

/***************************************************
 *	O Z C S 0001 1100
 *			POP
 ***************************************************/
static void f8_pop(void)
{
    ROMC_04();
}

/***************************************************
 *	O Z C S 0001 1101
 *	x x x x LR	 W,J
 ***************************************************/
static void f8_lr_w_j(void)
{
    ROMC_1C(cS);
    f8.w = f8.r[9]&0x1f;	//SKR mask needed to get Boris to work
}

/***************************************************
 *	O Z C S 0001 1110
 *	- - - - LR	 J,W
 ***************************************************/
static void f8_lr_j_w(void)
{
    f8.r[9] = f8.w&0x1f;	//SKR mask needed to get Boris to work
}

/***************************************************
 *	O Z C S 0001 1111
 *	x x x x INC
 ***************************************************/
static void f8_inc(void)
{
    CLR_OZCS;
    SET_OC(f8.a,1);
    f8.a += 1;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0000	aaaa aaaa
 *	- - - - LI	aa
 ***************************************************/
static void f8_li(void)
{
    ROMC_03(cL);
    f8.a = f8.dbus;
}

/***************************************************
 *	O Z C S 0010 0001	aaaa aaaa
 *	0 x 0 x NI	 aa
 ***************************************************/
static void f8_ni(void)
{
    ROMC_03(cL);
    CLR_OZCS;
    f8.a &= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0010	aaaa aaaa
 *	0 x 0 x OI	 aa
 ***************************************************/
static void f8_oi(void)
{
    ROMC_03(cL);
    CLR_OZCS;
    f8.a |= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0011	aaaa aaaa
 *	0 x 0 x XI	 aa
 ***************************************************/
static void f8_xi(void)
{
    ROMC_03(cL);
    CLR_OZCS;
    f8.a ^= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0100	aaaa aaaa
 *	x x x x AI	 aa
 ***************************************************/
static void f8_ai(void)
{
    ROMC_03(cL);
    CLR_OZCS;
    SET_OC(f8.a,f8.dbus);
    f8.a += f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0101	aaaa aaaa
 *	x x x x CI	 aa
 ***************************************************/
static void f8_ci(void)
{
    UINT16 tmp = ((UINT8)~f8.a) + 1;
    ROMC_03(cL);
    CLR_OZCS;
    SET_OC(tmp,f8.dbus);
    tmp += f8.dbus;
    SET_SZ((UINT8)tmp);
}

/***************************************************
 *	O Z C S 0010 0110	aaaa aaaa
 *	0 x 0 x IN	 aa
 ***************************************************/
static void f8_in(void)
{
    ROMC_03(cL);
    CLR_OZCS;
    ROMC_1B();
    f8.a = f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 0010 0111	aaaa aaaa
 *	- - - - OUT  aa
 ***************************************************/
static void f8_out(void)
{
    ROMC_03(cL);
    f8.dbus = f8.a;
    ROMC_1A();
}

/***************************************************
 *	O Z C S 0010 1000	iiii iiii	jjjj jjjj
 *	- - - - PI	 iijj
 ***************************************************/
static void f8_pi(void)
{
    ROMC_03(cL);
    f8.a = f8.dbus;
    ROMC_0D();
    ROMC_0C();
    f8.dbus = f8.a;
    ROMC_14();
}

/***************************************************
 *	O Z C S 0010 1001	iiii iiii	jjjj jjjj
 *	- - - - JMP  iijj
 ***************************************************/
static void f8_jmp(void)
{
    ROMC_03(cL);
    f8.a = f8.dbus;
    ROMC_0C();
    f8.dbus = f8.a;
    ROMC_14();
}

/***************************************************
 *	O Z C S 0010 1010	iiii iiii	jjjj jjjj
 *	- - - - DCI  iijj
 ***************************************************/
static void f8_dci(void)
{
    ROMC_11();
    ROMC_03(cS);
    ROMC_0E();
    ROMC_03(cS);
}

/***************************************************
 *	O Z C S 0010 1011
 *	- - - - NOP
 ***************************************************/
static void f8_nop(void)
{
}

/***************************************************
 *	O Z C S 0010 1100
 *	- - - - XDC
 ***************************************************/
static void f8_xdc(void)
{
    ROMC_1D();
}

/***************************************************
 *	O Z C S 0011 rrrr
 *	x x x x DS	 r
 ***************************************************/
static void f8_ds_r(int r)
{
    CLR_OZCS;
    SET_OC(f8.r[r], 0xff);
    f8.r[r] = f8.r[r] + 0xff;
    SET_SZ(f8.r[r]);
}

/***************************************************
 *	O Z C S 0011 1100
 *	x x x x DS	 ISAR
 ***************************************************/
static void f8_ds_isar(void)
{
    CLR_OZCS;
    SET_OC(f8.r[f8.is], 0xff);
    f8.r[f8.is] = f8.r[f8.is] + 0xff;
    SET_SZ(f8.r[f8.is]);
}

/***************************************************
 *	O Z C S 0011 1101
 *	x x x x DS	 ISAR++
 ***************************************************/
static void f8_ds_isar_i(void)
{
    CLR_OZCS;
    SET_OC(f8.r[f8.is], 0xff);
    f8.r[f8.is] = f8.r[f8.is] + 0xff;
    SET_SZ(f8.r[f8.is]);
    f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 0011 1110
 *	x x x x DS	ISAR--
 ***************************************************/
static void f8_ds_isar_d(void)
{
    CLR_OZCS;
    SET_OC(f8.r[f8.is], 0xff);
    f8.r[f8.is] = f8.r[f8.is] + 0xff;
    SET_SZ(f8.r[f8.is]);
    f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 0100 rrrr
 *	- - - - LR	A,r
 ***************************************************/
static void f8_lr_a_r(int r)
{
    f8.a = f8.r[r];
}

/***************************************************
 *	O Z C S 0100 1100
 *	- - - - LR	A,ISAR
 ***************************************************/
static void f8_lr_a_isar(void)
{
    f8.a = f8.r[f8.is];
}

/***************************************************
 *	O Z C S 0100 1101
 *	- - - - LR	A,ISAR++
 ***************************************************/
static void f8_lr_a_isar_i(void)
{
    f8.a = f8.r[f8.is];
    f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 0100 1110
 *	- - - - LR	A,ISAR--
 ***************************************************/
static void f8_lr_a_isar_d(void)
{
    f8.a = f8.r[f8.is];
    f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 0101 rrrr
 *	- - - - LR	r,A
 ***************************************************/
static void f8_lr_r_a(int r)
{
    f8.r[r] = f8.a;
}

/***************************************************
 *	O Z C S 0101 1100
 *	- - - - LR	ISAR,A
 ***************************************************/
static void f8_lr_isar_a(void)
{
    f8.r[f8.is] = f8.a;
}

/***************************************************
 *	O Z C S 0101 1101
 *	- - - - LR	ISAR++,A
 ***************************************************/
static void f8_lr_isar_i_a(void)
{
    f8.r[f8.is] = f8.a;
    f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 0101 1110
 *	- - - - LR	ISAR--,A
 ***************************************************/
static void f8_lr_isar_d_a(void)
{
    f8.r[f8.is] = f8.a;
    f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 0110 0eee
 *	- - - - LISU e
 ***************************************************/
static void f8_lisu(int e)
{
    f8.is = (f8.is & 0x07) | e;
}

/***************************************************
 *	O Z C S 0110 1eee
 *	- - - - LISL e
 ***************************************************/
static void f8_lisl(int e)
{
    f8.is = (f8.is & 0x38) | e;
}

/***************************************************
 *	O Z C S 0111 iiii
 *	- - - - LIS  i
 ***************************************************/
static void f8_lis(int i)
{
    f8.a = i;
}

/***************************************************
 *	O Z C S 1000 0eee	aaaa aaaa
 *			BT	 e,aa
 ***************************************************/
static void f8_bt(int e)
{
    ROMC_1C(cS);
    if (f8.w & e)
	ROMC_01();	   /* take the relative branch */
    else
	ROMC_03(cS);	   /* just read the argument on the data bus */
}

/***************************************************
 *	O Z C S 1000 1000
 *	x x x x AM
 ***************************************************/
static void f8_am(void)
{
    ROMC_02();
    CLR_OZCS;
    SET_OC(f8.a, f8.dbus);
    f8.a += f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1000 1001
 *	x x x x AMD
 ***************************************************/
static void f8_amd(void)
{
    UINT8 tmp = f8.a - 0x66, adj = 0x00;
    int sum;
    ROMC_02();
    sum = (tmp & 0x0f) + (f8.dbus & 0x0f);
    if (sum > 0x09)
	adj += 0x06;
    sum = tmp + f8.dbus + adj;
    if (sum > 0x99)
	adj += 0x60;
    tmp += adj;
    CLR_OZCS;
    SET_OC(tmp,f8.dbus);
    f8.a = tmp + f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1000 1010
 *	0 x 0 x NM
 ***************************************************/
static void f8_nm(void)
{
    ROMC_02();
    CLR_OZCS;
    f8.a &= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1000 1011
 *	0 x 0 x OM
 ***************************************************/
static void f8_om(void)
{
    ROMC_02();
    CLR_OZCS;
    f8.a |= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1000 1100
 *	0 x 0 x XM
 ***************************************************/
static void f8_xm(void)
{
    ROMC_02();
    CLR_OZCS;
    f8.a ^= f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1000 1101
 *	0 x 0 x CM
 ***************************************************/
static void f8_cm(void)	/* SKR changed to match f8_ci() */
{
    UINT16 tmp = ((UINT8)~f8.a) + 1;
	ROMC_02();
	CLR_OZCS;
	SET_OC(tmp,f8.dbus);
	tmp += f8.dbus;
	SET_SZ((UINT8)tmp);
}

/***************************************************
 *	O Z C S 1000 1110
 *	- - - - ADC
 ***************************************************/
static void f8_adc(void)
{
    f8.dbus = f8.a;
    ROMC_0A();			/* add data bus value to DC0 */
}

/***************************************************
 *	O Z C S 1000 1111
 *	- - - - BR7
 ***************************************************/
static void f8_br7(void)
{
    if ((f8.is & 7) == 7)
	ROMC_03(cS);		/* just read the argument on the data bus */
    else
	ROMC_01();		/* take the relative branch */
}

/***************************************************
 *	O Z C S 1001 tttt	aaaa aaaa
 *	- - - - BF	 t,aa
 ***************************************************/
static void f8_bf(int t)
{
    ROMC_1C(cS);
    if (f8.w & t)
        ROMC_03(cS);      /* just read the argument on the data bus */
    else
        ROMC_01();      /* take the relative branch */
}

/***************************************************
 *	O Z C S 1010 000n
 *	0 x 0 x INS  n				(n = 0-1)
 ***************************************************/
static void f8_ins_0(int n)
{
    ROMC_1C(cS);
    CLR_OZCS;
    f8.a = io_read_byte_8(n,&f8);
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1010 nnnn
 *	0 x 0 x INS  n				(n = 4-F)
 ***************************************************/
static void f8_ins_1(int n)
{
    ROMC_1C(cL);
    f8.io = n;
    ROMC_1B();
    CLR_OZCS;
    f8.a = f8.dbus;
    SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1011 000n
 *	- - - - OUTS n				(n = 0-1)
 ***************************************************/
static void f8_outs_0(int n)
{
    ROMC_1C(cS);
    io_write_byte_8(n, f8.a,&f8);
}

/***************************************************
 *	O Z C S 1011 nnnn
 *	- - - - OUTS n				(n = 4-F)
 ***************************************************/
static void f8_outs_1(int n)
{
    ROMC_1C(cL);
    f8.io = n;
    f8.dbus = f8.a;
    ROMC_1A();
}

/***************************************************
 *	O Z C S 1100 rrrr
 *	x x x x AS	 r
 ***************************************************/
static void f8_as(int r)
{
	CLR_OZCS;
	SET_OC(f8.a, f8.r[r]);
	f8.a += f8.r[r];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1100 1100
 *	x x x x AS	 ISAR
 ***************************************************/
static void f8_as_isar(void)
{
	CLR_OZCS;
	SET_OC(f8.a, f8.r[f8.is]);
	f8.a += f8.r[f8.is];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1100 1101
 *	x x x x AS	 ISAR++
 ***************************************************/
static void f8_as_isar_i(void)
{
	CLR_OZCS;
	SET_OC(f8.a, f8.r[f8.is]);
	f8.a += f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 1100 1110
 *	x x x x AS	 ISAR--
 ***************************************************/
static void f8_as_isar_d(void)
{
	CLR_OZCS;
	SET_OC(f8.a, f8.r[f8.is]);
	f8.a += f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 1101 rrrr
 *	x x x x ASD  r
 ***************************************************/
static void f8_asd(int r)
{
	UINT8 tmp = f8.a - 0x66, adj = 0x00;
	int sum;
	ROMC_1C(cS);
	sum = (tmp & 0x0f) + (f8.r[r] & 0x0f);
	if (sum > 0x09)
		adj += 0x06;
	sum = tmp + f8.r[r] + adj;
	if (sum > 0x99)
		adj += 0x60;
	tmp += adj;
	CLR_OZCS;
	SET_OC(tmp, f8.r[r]);
	f8.a = tmp + f8.r[r];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1101 1100
 *	x x x x ASD  ISAR
 ***************************************************/
static void f8_asd_isar(void)
{
	UINT8 tmp = f8.a - 0x66, adj = 0x00;
	int sum;
	ROMC_1C(cS);
	sum = (tmp & 0x0f) + (f8.r[f8.is] & 0x0f);
	if (sum > 0x09)
		adj += 0x06;
	sum = tmp + f8.r[f8.is] + adj;
	if (sum > 0x99)
		adj += 0x60;
	tmp += adj;
	CLR_OZCS;
	SET_OC(tmp, f8.r[f8.is]);
	f8.a = tmp + f8.r[f8.is];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1101 1101
 *	x x x x ASD  ISAR++
 ***************************************************/
static void f8_asd_isar_i(void)
{
	UINT8 tmp = f8.a - 0x66, adj = 0x00;
	int sum;
	ROMC_1C(cS);
	sum = (tmp & 0x0f) + (f8.r[f8.is] & 0x0f);
	if (sum > 0x09)
		adj += 0x06;
	sum = tmp + f8.r[f8.is] + adj;
	if (sum > 0x99)
		adj += 0x60;
	tmp += adj;
	CLR_OZCS;
	SET_OC(tmp, f8.r[f8.is]);
	f8.a = tmp + f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 1101 1110
 *	x x x x ASD  ISAR--
 ***************************************************/
static void f8_asd_isar_d(void)
{
	UINT8 tmp = f8.a - 0x66, adj = 0x00;
	int sum;
	ROMC_1C(cS);
	sum = (tmp & 0x0f) + (f8.r[f8.is] & 0x0f);
	if (sum > 0x09)
		adj += 0x06;
	sum = tmp + f8.r[f8.is] + adj;
	if (sum > 0x99)
		adj += 0x60;
	tmp += adj;
	CLR_OZCS;
	SET_OC(tmp, f8.r[f8.is]);
	f8.a = tmp + f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 1110 rrrr
 *	0 x 0 x XS	 r
 ***************************************************/
static void f8_xs(int r)
{
	CLR_OZCS;
	f8.a ^= f8.r[r];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1110 1100
 *	0 x 0 x XS	 ISAR
 ***************************************************/
static void f8_xs_isar(void)
{
	CLR_OZCS;
	f8.a ^= f8.r[f8.is];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1110 1101
 *	0 x 0 x XS	 ISAR++
 ***************************************************/
static void f8_xs_isar_i(void)
{
	CLR_OZCS;
	f8.a ^= f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 1110 1110
 *	0 x 0 x XS	 ISAR--
 ***************************************************/
static void f8_xs_isar_d(void)
{
	CLR_OZCS;
	f8.a ^= f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

/***************************************************
 *	O Z C S 1111 rrrr
 *	0 x 0 x NS	 r
 ***************************************************/
static void f8_ns(int r)
{
	CLR_OZCS;
	f8.a &= f8.r[r];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1111 1100
 *	0 x 0 x NS	 ISAR
 ***************************************************/
static void f8_ns_isar(void)
{
	CLR_OZCS;
	f8.a &= f8.r[f8.is];
	SET_SZ(f8.a);
}

/***************************************************
 *	O Z C S 1111 1101
 *	0 x 0 x NS	 ISAR++
 ***************************************************/
static void f8_ns_isar_i(void)
{
	CLR_OZCS;
	f8.a &= f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is + 1) & 0x07);
}

/***************************************************
 *	O Z C S 1111 1110
 *	0 x 0 x NS	 ISAR--
 ***************************************************/
static void f8_ns_isar_d(void)
{
	CLR_OZCS;
	f8.a &= f8.r[f8.is];
	SET_SZ(f8.a);
	f8.is = (f8.is & 0x38) | ((f8.is - 1) & 0x07);
}

void f8_reset(void *param)
{
	System::Diagnostics::Debug::WriteLine("reset");

//	memset(&f8, 0, sizeof(f8_Regs));
	f8.pc0=0;
	f8.pc1=0;
	f8.dc0=0;
	f8.dc1=0;
	f8.a=0;
	f8.w=0;
	f8.is=0;
	f8.dbus=0;
	f8.io=0;
	f8.irq_vector=0x20;
	f8.irq_request=0;
	for(int i=0;i<64;i++)
		f8.r[i]=0;

	f8.oldpc0=1;

	f8.timercount=0;
	f8.timersubcount=0;
	f8.timerrunning=0;
	f8.timerintenabled=0;
	f8.timerdivisor=1;
	f8.timermodulon=255;
	f8.timerirq=0;

	f8.w&=~W_I;	//clear ICB

	/* save PC0 to PC1 and reset PC0 */
	ROMC_08();
	/* fetch the first opcode */
	ROMC_00(cS);

	//ports are probably reset
	for(int i=0;i<8;i++)
		olddigs[i]=0;
}

void f8_execute(StreamWriter^ sw) {

	char *buffer=(char *)malloc(256);
	String^ sbuffer;

	UINT8 op=f8.dbus;

	if(show||singlestep) {
		f8dasm(rom,f8.oldpc0-1,buffer);
		sbuffer=gcnew String(buffer);
	}

	switch( op ) {
	/* opcode  bitmask */
	case 0x00: /* 0000 0000 */	f8_lr_a_ku();		break;
	case 0x01: /* 0000 0001 */	f8_lr_a_kl();		break;
	case 0x02: /* 0000 0010 */	f8_lr_a_qu();		break;
	case 0x03: /* 0000 0011 */	f8_lr_a_ql();		break;
	case 0x04: /* 0000 0100 */	f8_lr_ku_a();		break;
	case 0x05: /* 0000 0101 */	f8_lr_kl_a();		break;
	case 0x06: /* 0000 0110 */	f8_lr_qu_a();		break;
	case 0x07: /* 0000 0111 */	f8_lr_ql_a();		break;
	case 0x08: /* 0000 1000 */	f8_lr_k_p();		break;
	case 0x09: /* 0000 1001 */	f8_lr_p_k();		break;
	case 0x0a: /* 0000 1010 */	f8_lr_a_is();		break;
	case 0x0b: /* 0000 1011 */	f8_lr_is_a();		break;
	case 0x0c: /* 0000 1100 */	f8_pk();			break;
	case 0x0d: /* 0000 1101 */	f8_lr_p0_q();		break;
	case 0x0e: /* 0000 1110 */	f8_lr_q_dc();		break;
	case 0x0f: /* 0000 1111 */	f8_lr_dc_q();		break;

    case 0x10: /* 0001 0000 */  f8_lr_dc_h();       break;
	case 0x11: /* 0001 0001 */	f8_lr_h_dc();		break;
	case 0x12: /* 0001 0010 */	f8_sr_1();			break;
	case 0x13: /* 0001 0011 */	f8_sl_1();			break;
	case 0x14: /* 0001 0100 */	f8_sr_4();			break;
	case 0x15: /* 0001 0101 */	f8_sl_4();			break;
	case 0x16: /* 0001 0110 */	f8_lm();			break;
	case 0x17: /* 0001 0111 */	f8_st();			break;
	case 0x18: /* 0001 1000 */	f8_com();			break;
	case 0x19: /* 0001 1001 */	f8_lnk();			break;

	case 0x1a: /* 0001 1010 */	f8_di();			
#if DEBUGFILE
		sw->WriteLine("DI");
#endif
		break;
	case 0x1b: /* 0001 1011 */	f8_ei();			
#if DEBUGFILE
		sw->WriteLine("EI");
#endif
		break;

	case 0x1c: /* 0001 1100 */	f8_pop();			break;
	case 0x1d: /* 0001 1101 */	f8_lr_w_j();		break;
	case 0x1e: /* 0001 1110 */	f8_lr_j_w();		break;
	case 0x1f: /* 0001 1111 */	f8_inc();			break;

    case 0x20: /* 0010 0000 */  f8_li();            break;
	case 0x21: /* 0010 0001 */	f8_ni();			break;
	case 0x22: /* 0010 0010 */	f8_oi();			break;
	case 0x23: /* 0010 0011 */	f8_xi();			break;
	case 0x24: /* 0010 0100 */	f8_ai();			break;
	case 0x25: /* 0010 0101 */	f8_ci();			break;
	case 0x26: /* 0010 0110 */	f8_in();			break;
	case 0x27: /* 0010 0111 */	f8_out();			break;
	case 0x28: /* 0010 1000 */	f8_pi();			break;
	case 0x29: /* 0010 1001 */	f8_jmp();			break;
	case 0x2a: /* 0010 1010 */	f8_dci();			break;
	case 0x2b: /* 0010 1011 */	f8_nop();			break;
	case 0x2c: /* 0010 1100 */	f8_xdc();			break;
	case 0x2d: /* 0010 1101 */	illegal();			break;
	case 0x2e: /* 0010 1110 */	illegal();			break;
	case 0x2f: /* 0010 1111 */	illegal();			break;

    case 0x30: /* 0011 0000 */  f8_ds_r( 0);        break;
	case 0x31: /* 0011 0001 */	f8_ds_r( 1);		break;
	case 0x32: /* 0011 0010 */	f8_ds_r( 2);		break;
	case 0x33: /* 0011 0011 */	f8_ds_r( 3);		break;
	case 0x34: /* 0011 0100 */	f8_ds_r( 4);		break;
	case 0x35: /* 0011 0101 */	f8_ds_r( 5);		break;
	case 0x36: /* 0011 0110 */	f8_ds_r( 6);		break;
	case 0x37: /* 0011 0111 */	f8_ds_r( 7);		break;
	case 0x38: /* 0011 1000 */	f8_ds_r( 8);		break;
	case 0x39: /* 0011 1001 */	f8_ds_r( 9);		break;
	case 0x3a: /* 0011 1010 */	f8_ds_r(10);		break;
	case 0x3b: /* 0011 1011 */	f8_ds_r(11);		break;
	case 0x3c: /* 0011 1100 */	f8_ds_isar();		break;
	case 0x3d: /* 0011 1101 */	f8_ds_isar_i(); 	break;
	case 0x3e: /* 0011 1110 */	f8_ds_isar_d(); 	break;
	case 0x3f: /* 0011 1111 */	illegal();			break;

    case 0x40: /* 0100 0000 */  f8_lr_a_r( 0);      break;
	case 0x41: /* 0100 0001 */	f8_lr_a_r( 1);		break;
	case 0x42: /* 0100 0010 */	f8_lr_a_r( 2);		break;
	case 0x43: /* 0100 0011 */	f8_lr_a_r( 3);		break;
	case 0x44: /* 0100 0100 */	f8_lr_a_r( 4);		break;
	case 0x45: /* 0100 0101 */	f8_lr_a_r( 5);		break;
	case 0x46: /* 0100 0110 */	f8_lr_a_r( 6);		break;
	case 0x47: /* 0100 0111 */	f8_lr_a_r( 7);		break;
	case 0x48: /* 0100 1000 */	f8_lr_a_r( 8);		break;
	case 0x49: /* 0100 1001 */	f8_lr_a_r( 9);		break;
	case 0x4a: /* 0100 1010 */	f8_lr_a_r(10);		break;
	case 0x4b: /* 0100 1011 */	f8_lr_a_r(11);		break;
	case 0x4c: /* 0100 1100 */	f8_lr_a_isar(); 	break;
	case 0x4d: /* 0100 1101 */	f8_lr_a_isar_i();	break;
	case 0x4e: /* 0100 1110 */	f8_lr_a_isar_d();	break;
	case 0x4f: /* 0100 1111 */	illegal();			break;

	case 0x50: /* 0101 0000 */	f8_lr_r_a( 0);		break;
	case 0x51: /* 0101 0001 */	f8_lr_r_a( 1);		break;
	case 0x52: /* 0101 0010 */	f8_lr_r_a( 2);		break;
	case 0x53: /* 0101 0011 */	f8_lr_r_a( 3);		break;
	case 0x54: /* 0101 0100 */	f8_lr_r_a( 4);		break;
	case 0x55: /* 0101 0101 */	f8_lr_r_a( 5);		break;
	case 0x56: /* 0101 0110 */	f8_lr_r_a( 6);		break;
	case 0x57: /* 0101 0111 */	f8_lr_r_a( 7);		break;
	case 0x58: /* 0101 1000 */	f8_lr_r_a( 8);		break;
	case 0x59: /* 0101 1001 */	f8_lr_r_a( 9);		break;
	case 0x5a: /* 0101 1010 */	f8_lr_r_a(10);		break;
	case 0x5b: /* 0101 1011 */	f8_lr_r_a(11);		break;
	case 0x5c: /* 0101 1100 */	f8_lr_isar_a(); 	break;
	case 0x5d: /* 0101 1101 */	f8_lr_isar_i_a();	break;
	case 0x5e: /* 0101 1110 */	f8_lr_isar_d_a();	break;
	case 0x5f: /* 0101 1111 */	illegal();			break;

	case 0x60: /* 0110 0000 */	f8_lisu(0x00);		break;
	case 0x61: /* 0110 0001 */	f8_lisu(0x08);		break;
	case 0x62: /* 0110 0010 */	f8_lisu(0x10);		break;
	case 0x63: /* 0110 0011 */	f8_lisu(0x18);		break;
	case 0x64: /* 0110 0100 */	f8_lisu(0x20);		break;
	case 0x65: /* 0110 0101 */	f8_lisu(0x28);		break;
	case 0x66: /* 0110 0110 */	f8_lisu(0x30);		break;
	case 0x67: /* 0110 0111 */	f8_lisu(0x38);		break;
	case 0x68: /* 0110 1000 */	f8_lisl(0x00);		break;
	case 0x69: /* 0110 1001 */	f8_lisl(0x01);		break;
	case 0x6a: /* 0110 1010 */	f8_lisl(0x02);		break;
	case 0x6b: /* 0110 1011 */	f8_lisl(0x03);		break;
	case 0x6c: /* 0110 1100 */	f8_lisl(0x04);		break;
	case 0x6d: /* 0110 1101 */	f8_lisl(0x05);		break;
	case 0x6e: /* 0110 1110 */	f8_lisl(0x06);		break;
	case 0x6f: /* 0110 1111 */	f8_lisl(0x07);		break;

	case 0x70: /* 0111 0000 */	f8_lis(0x0);		break;
	case 0x71: /* 0111 0001 */	f8_lis(0x1);		break;
	case 0x72: /* 0111 0010 */	f8_lis(0x2);		break;
	case 0x73: /* 0111 0011 */	f8_lis(0x3);		break;
	case 0x74: /* 0111 0100 */	f8_lis(0x4);		break;
	case 0x75: /* 0111 0101 */	f8_lis(0x5);		break;
	case 0x76: /* 0111 0110 */	f8_lis(0x6);		break;
	case 0x77: /* 0111 0111 */	f8_lis(0x7);		break;
	case 0x78: /* 0111 1000 */	f8_lis(0x8);		break;
	case 0x79: /* 0111 1001 */	f8_lis(0x9);		break;
	case 0x7a: /* 0111 1010 */	f8_lis(0xa);		break;
	case 0x7b: /* 0111 1011 */	f8_lis(0xb);		break;
	case 0x7c: /* 0111 1100 */	f8_lis(0xc);		break;
	case 0x7d: /* 0111 1101 */	f8_lis(0xd);		break;
	case 0x7e: /* 0111 1110 */	f8_lis(0xe);		break;
	case 0x7f: /* 0111 1111 */	f8_lis(0xf);		break;

	case 0x80: /* 1000 0000 */	f8_bt(0);			break;
	case 0x81: /* 1000 0001 */	f8_bt(1);			break;
	case 0x82: /* 1000 0010 */	f8_bt(2);			break;
	case 0x83: /* 1000 0011 */	f8_bt(3);			break;
	case 0x84: /* 1000 0100 */	f8_bt(4);			break;
	case 0x85: /* 1000 0101 */	f8_bt(5);			break;
	case 0x86: /* 1000 0110 */	f8_bt(6);			break;
	case 0x87: /* 1000 0111 */	f8_bt(7);			break;
	case 0x88: /* 1000 1000 */	f8_am();			break;
	case 0x89: /* 1000 1001 */	f8_amd();			break;
	case 0x8a: /* 1000 1010 */	f8_nm();			break;
	case 0x8b: /* 1000 1011 */	f8_om();			break;
	case 0x8c: /* 1000 1100 */	f8_xm();			break;
	case 0x8d: /* 1000 1101 */	f8_cm();			break;
	case 0x8e: /* 1000 1110 */	f8_adc();			break;
	case 0x8f: /* 1000 1111 */	f8_br7();			break;

	case 0x90: /* 1001 0000 */	f8_bf(0x0); 		break;
	case 0x91: /* 1001 0001 */	f8_bf(0x1); 		break;
	case 0x92: /* 1001 0010 */	f8_bf(0x2); 		break;
	case 0x93: /* 1001 0011 */	f8_bf(0x3); 		break;
	case 0x94: /* 1001 0100 */	f8_bf(0x4); 		break;
	case 0x95: /* 1001 0101 */	f8_bf(0x5); 		break;
	case 0x96: /* 1001 0110 */	f8_bf(0x6); 		break;
	case 0x97: /* 1001 0111 */	f8_bf(0x7); 		break;
	case 0x98: /* 1001 1000 */	f8_bf(0x8); 		break;
	case 0x99: /* 1001 1001 */	f8_bf(0x9); 		break;
	case 0x9a: /* 1001 1010 */	f8_bf(0xa); 		break;
	case 0x9b: /* 1001 1011 */	f8_bf(0xb); 		break;
	case 0x9c: /* 1001 1100 */	f8_bf(0xc); 		break;
	case 0x9d: /* 1001 1101 */	f8_bf(0xd); 		break;
	case 0x9e: /* 1001 1110 */	f8_bf(0xe); 		break;
	case 0x9f: /* 1001 1111 */	f8_bf(0xf); 		break;

	case 0xa0: /* 1010 0000 */	f8_ins_0(0x0);		break;
	case 0xa1: /* 1010 0001 */	f8_ins_0(0x1);		break;
	case 0xa2: /* 1010 0010 */	illegal();			break;
	case 0xa3: /* 1010 0011 */	illegal();			break;
	case 0xa4: /* 1010 0100 */	f8_ins_1(0x4);		break;
	case 0xa5: /* 1010 0101 */	f8_ins_1(0x5);		break;
	case 0xa6: /* 1010 0110 */	f8_ins_1(0x6);		break;
	case 0xa7: /* 1010 0111 */	f8_ins_1(0x7);		break;
	case 0xa8: /* 1010 1000 */	f8_ins_1(0x8);		break;
	case 0xa9: /* 1010 1001 */	f8_ins_1(0x9);		break;
	case 0xaa: /* 1010 1010 */	f8_ins_1(0xa);		break;
	case 0xab: /* 1010 1011 */	f8_ins_1(0xb);		break;
	case 0xac: /* 1010 1100 */	f8_ins_1(0xc);		break;
	case 0xad: /* 1010 1101 */	f8_ins_1(0xd);		break;
	case 0xae: /* 1010 1110 */	f8_ins_1(0xe);		break;
	case 0xaf: /* 1010 1111 */	f8_ins_1(0xf);		break;

	case 0xb0: /* 1011 0000 */	f8_outs_0(0x0); 	break;
	case 0xb1: /* 1011 0001 */	f8_outs_0(0x1); 	break;
	case 0xb2: /* 1011 0010 */	illegal();			break;
	case 0xb3: /* 1011 0011 */	illegal();			break;
	case 0xb4: /* 1011 0100 */	f8_outs_1(0x4); 	break;
	case 0xb5: /* 1011 0101 */	f8_outs_1(0x5); 	break;
	case 0xb6: /* 1011 0110 */	f8_outs_1(0x6); 	break;
	case 0xb7: /* 1011 0111 */	f8_outs_1(0x7); 	break;
	case 0xb8: /* 1011 1000 */	f8_outs_1(0x8); 	break;
	case 0xb9: /* 1011 1001 */	f8_outs_1(0x9); 	break;
	case 0xba: /* 1011 1010 */	f8_outs_1(0xa); 	break;
	case 0xbb: /* 1011 1011 */	f8_outs_1(0xb); 	break;
	case 0xbc: /* 1011 1100 */	f8_outs_1(0xc); 	break;
	case 0xbd: /* 1011 1101 */	f8_outs_1(0xd); 	break;
	case 0xbe: /* 1011 1110 */	f8_outs_1(0xe); 	break;
	case 0xbf: /* 1011 1111 */	f8_outs_1(0xf); 	break;

	case 0xc0: /* 1100 0000 */	f8_as(0x0); 		break;
	case 0xc1: /* 1100 0001 */	f8_as(0x1); 		break;
	case 0xc2: /* 1100 0010 */	f8_as(0x2); 		break;
	case 0xc3: /* 1100 0011 */	f8_as(0x3); 		break;
	case 0xc4: /* 1100 0100 */	f8_as(0x4); 		break;
	case 0xc5: /* 1100 0101 */	f8_as(0x5); 		break;
	case 0xc6: /* 1100 0110 */	f8_as(0x6); 		break;
	case 0xc7: /* 1100 0111 */	f8_as(0x7); 		break;
	case 0xc8: /* 1100 1000 */	f8_as(0x8); 		break;
	case 0xc9: /* 1100 1001 */	f8_as(0x9); 		break;
	case 0xca: /* 1100 1010 */	f8_as(0xa); 		break;
	case 0xcb: /* 1100 1011 */	f8_as(0xb); 		break;
	case 0xcc: /* 1100 1100 */	f8_as_isar(); 		break;
	case 0xcd: /* 1100 1101 */	f8_as_isar_i(); 	break;
	case 0xce: /* 1100 1110 */	f8_as_isar_d(); 	break;
	case 0xcf: /* 1100 1111 */	illegal(); 			break;

	case 0xd0: /* 1101 0000 */	f8_asd(0x0);		break;
	case 0xd1: /* 1101 0001 */	f8_asd(0x1);		break;
	case 0xd2: /* 1101 0010 */	f8_asd(0x2);		break;
	case 0xd3: /* 1101 0011 */	f8_asd(0x3);		break;
	case 0xd4: /* 1101 0100 */	f8_asd(0x4);		break;
	case 0xd5: /* 1101 0101 */	f8_asd(0x5);		break;
	case 0xd6: /* 1101 0110 */	f8_asd(0x6);		break;
	case 0xd7: /* 1101 0111 */	f8_asd(0x7);		break;
	case 0xd8: /* 1101 1000 */	f8_asd(0x8);		break;
	case 0xd9: /* 1101 1001 */	f8_asd(0x9);		break;
	case 0xda: /* 1101 1010 */	f8_asd(0xa);		break;
	case 0xdb: /* 1101 1011 */	f8_asd(0xb);		break;
	case 0xdc: /* 1101 1100 */	f8_asd_isar();		break;
	case 0xdd: /* 1101 1101 */	f8_asd_isar_i();	break;
	case 0xde: /* 1101 1110 */	f8_asd_isar_d();	break;
	case 0xdf: /* 1101 1111 */	illegal();			break;

	case 0xe0: /* 1110 0000 */	f8_xs(0x0); 		break;
	case 0xe1: /* 1110 0001 */	f8_xs(0x1); 		break;
	case 0xe2: /* 1110 0010 */	f8_xs(0x2); 		break;
	case 0xe3: /* 1110 0011 */	f8_xs(0x3); 		break;
	case 0xe4: /* 1110 0100 */	f8_xs(0x4); 		break;
	case 0xe5: /* 1110 0101 */	f8_xs(0x5); 		break;
	case 0xe6: /* 1110 0110 */	f8_xs(0x6); 		break;
	case 0xe7: /* 1110 0111 */	f8_xs(0x7); 		break;
	case 0xe8: /* 1110 1000 */	f8_xs(0x8); 		break;
	case 0xe9: /* 1110 1001 */	f8_xs(0x9); 		break;
	case 0xea: /* 1110 1010 */	f8_xs(0xa); 		break;
	case 0xeb: /* 1110 1011 */	f8_xs(0xb); 		break;
	case 0xec: /* 1110 1100 */	f8_xs_isar();		break;
	case 0xed: /* 1110 1101 */	f8_xs_isar_i();		break;
	case 0xee: /* 1110 1110 */	f8_xs_isar_d();		break;
	case 0xef: /* 1110 1111 */	illegal();			break;

	case 0xf0: /* 1111 0000 */	f8_ns(0x0); 		break;
	case 0xf1: /* 1111 0001 */	f8_ns(0x1); 		break;
	case 0xf2: /* 1111 0010 */	f8_ns(0x2); 		break;
	case 0xf3: /* 1111 0011 */	f8_ns(0x3); 		break;
	case 0xf4: /* 1111 0100 */	f8_ns(0x4); 		break;
	case 0xf5: /* 1111 0101 */	f8_ns(0x5); 		break;
	case 0xf6: /* 1111 0110 */	f8_ns(0x6); 		break;
	case 0xf7: /* 1111 0111 */	f8_ns(0x7); 		break;
	case 0xf8: /* 1111 1000 */	f8_ns(0x8); 		break;
	case 0xf9: /* 1111 1001 */	f8_ns(0x9); 		break;
	case 0xfa: /* 1111 1010 */	f8_ns(0xa); 		break;
	case 0xfb: /* 1111 1011 */	f8_ns(0xb); 		break;
	case 0xfc: /* 1111 1100 */	f8_ns_isar();		break;
	case 0xfd: /* 1111 1101 */	f8_ns_isar_i();		break;
	case 0xfe: /* 1111 1110 */	f8_ns_isar_d();		break;
	case 0xff: /* 1111 1111 */	illegal();			break;
    }	//switch(op)


	if(f8.timerirq&&f8.timerintenabled) {		//timercount has rolled over and the timer is enabled
#if DEBUGFILE
		sw->WriteLine("timer IRQ");
#endif
		f8.irq_request=1;	//simulate the irq
	}

	//don't check for IRQ after these instructions - SKR I think it should be 0C:PK, not 0D:LR P0,Q; also, 27:OUT only if not port 0/1
	if(op==0x0c||op==0x1b||op==0x1c||op==0x1d||op==0x28||op==0x29||(op>=0xb4&&op<=0xbf)||(op==0x27&&f8.io>1)) {
	    ROMC_00(cS);
	} else {	//check for IRQ after any other instruction
	    if (f8.w&W_I && f8.irq_request) {	//if ICB set and request has occurred

#if DEBUGFILE
			sw->WriteLine("f8 IRQ");
#endif
			ROMC_1C(cL);
			ROMC_0F();
			ROMC_13();
	    }

		if((op>=0x30)&&(op<=0x3f))	/* SKR - DS is a long cycle inst */
	    	ROMC_00(cL);
	    else
		    ROMC_00(cS);
	}

	if(show||singlestep) {
		Console::WriteLine(String::Concat("$",(f8.oldpc0-1).ToString("X4"),":$",op.ToString("X2")," ",sbuffer," Acc:$",f8.a.ToString("X2")," Flags:$",f8.w.ToString("X2")," ",
			f8.w & W_I ? "I":".",
			f8.w & W_O ? "O":".",
			f8.w & W_Z ? "Z":".",
			f8.w & W_C ? "C":".",
			f8.w & W_S ? "S":".",
			" ISAR:$",f8.is.ToString("X2"),
			" DC0:$",f8.dc0.ToString("X4"),
			" [DC0]:$",rom[(f8.dc0)&ADDRESSMASK].ToString("X2")
			));
	}

#if DEBUGFILE
		sw->WriteLine(String::Concat("$",(f8.oldpc0-1).ToString("X4"),":$",op.ToString("X2")," ",sbuffer," Acc:$",f8.a.ToString("X2")," Flags:$",f8.w.ToString("X2")," ",
			f8.w & W_I ? "I":".",
			f8.w & W_O ? "O":".",
			f8.w & W_Z ? "Z":".",
			f8.w & W_C ? "C":".",
			f8.w & W_S ? "S":".",
			" ISAR:$",f8.is.ToString("X2"),
			" DC0:$",f8.dc0.ToString("X4"),
			" [DC0]:$",rom[(f8.dc0)&ADDRESSMASK].ToString("X2")
			));
#endif

	free(buffer);
	f8.oldpc0=f8.pc0;
}	//f8_execute


void OnCancelKeyPressed(Object^ sender, ConsoleCancelEventArgs^ args) {
    args->Cancel = true;
	singlestep=true;
}

String^ piece(UINT8 romdata) {
	String^ piecename="?? ";
	if((romdata&0xff)==0x00)
		piecename=".. ";
	else if((romdata&0xff)==0x01)
		piecename="WP ";
	else if((romdata&0xff)==0x04)
		piecename="WR ";
	else if((romdata&0xff)==0x02)
		piecename="WN ";
	else if((romdata&0xff)==0x03)
		piecename="WB ";
	else if((romdata&0xff)==0x05)
		piecename="WQ ";
	else if((romdata&0xff)==0x06)
		piecename="WK ";
	else if((romdata&0xff)==0xff)
		piecename="BP ";
	else if((romdata&0xff)==0xfc)
		piecename="BR ";
	else if((romdata&0xff)==0xfe)
		piecename="BN ";
	else if((romdata&0xff)==0xfd)
		piecename="BB ";
	else if((romdata&0xff)==0xfb)
		piecename="BQ ";
	else if((romdata&0xff)==0xfa)
		piecename="BK ";

	return piecename;
}

int main(array<System::String ^> ^args)
{
	FileStream^ fs;
	StreamWriter^ sw=nullptr;
	FileInfo^ fi;

#if DEBUGFILE
	fi = gcnew FileInfo( "f8debug.txt" );
	fi->Delete();
	fs=gcnew FileStream("f8debug.txt",FileMode::OpenOrCreate);
	sw=gcnew StreamWriter(fs);
#endif

#if DEBUGPC0
	fi = gcnew FileInfo( "f8pc0.txt" );
	fi->Delete();
	FileStream^ pfs=gcnew FileStream("f8pc0.txt",FileMode::OpenOrCreate);
	StreamWriter^ psw=gcnew StreamWriter(pfs);
#endif

	Console::WriteLine("Boris Diplomat (brown) emulator based on Juergen Buchmueller's F8 emulator");

//	try {
		// Turn off the default system behavior when CTRL+C is pressed. When 
		// Console.TreatControlCAsInput is false, CTRL+C is treated as an
		// interrupt instead of as input.
		Console::TreatControlCAsInput = false;

		// Establish an event handler to process key press events.
		Console::CancelKeyPress +=  gcnew ConsoleCancelEventHandler(OnCancelKeyPressed);

		UINT8 lastbp=0xff;
		show=false;
		keycol=0;
		keyrow=9;
		keydown=0;
	
		//hard coded breakpoints for testing
		if(NUMBPS>0)
			bps[0]=0x50e0;	//CE
		if(NUMBPS>1)
			bps[1]=0x510f;	//RANK
		if(NUMBPS>2)
			bps[2]=0x5121;	//ENTER
		if(NUMBPS>3)
			bps[3]=0x515e;	//TIME/RANK

		for(int i=0;i<8;i++)
			olddigs[i]=0;

		bool doit=true;

		//load hex file
		try {
			FileStream^ fs = gcnew FileStream("boris.bin", FileMode::Open,System::IO::FileAccess::Read,System::IO::FileShare::None,2048);//specify 2K buffer
			BinaryReader^ breader;
			breader=gcnew BinaryReader(fs);
			array<Byte>^lrom=breader->ReadBytes(ROMSIZE);
			for(int i=0;i<ROMSIZE;i++) {
				rom[i]=lrom[i];
				access[i]=0;
				numaccesses[i]=0;
			}
			breader->Close();
			fs->Close();
		} catch(Exception^) {
			Console::WriteLine("Error loading boris.bin ROM file");
			Console::ReadLine();
			doit=false;
		}

		f8_reset(0);

		//show LEDs at powerup
		for(int i=0;i<8;i++)
			showdig(digits[i]);
		Console::WriteLine("");
	
		String^ line;

		while(doit) {
			bool doex=true;

#if DEBUGPC0
			psw->WriteLine("{0:X4}",f8.pc0-1);
#endif

			//breakpoint stuff for testing
			bool hitbp=false;
			for(int i=0;i<NUMBPS;i++) {
				if(((f8.pc0-1)&ADDRESSMASK)==(bps[i]&ADDRESSMASK)) {
					singlestep=true;
					if(lastbp!=i)
						Console::WriteLine(String::Concat("bp hit:",bps[i].ToString("X2")));
					lastbp=i;
					hitbp=true;
				}
			}
			if(!hitbp)
				lastbp=0xff;

			//output LEDs if different from last time
			bool updateLEDs=false;
			for(int i=0;i<8;i++) {
				if(digits[i]!=olddigs[i])
					updateLEDs=true;
				olddigs[i]=digits[i];
			}
			if(updateLEDs) {
				for(int i=0;i<8;i++)
					showdig(digits[i]);
				Console::WriteLine("");
			}

			//get input, if any
			if(Console::KeyAvailable)
				line=Console::ReadLine();

			if(singlestep) {
				Console::WriteLine("?");
				line=Console::ReadLine();
			}
			//decode input
			if(line!=nullptr) {
				doex=true;
				if(line->Length==0) {
					doex=false;
				} else if(!String::Compare(line,"help",true)) {
					Console::WriteLine("quit - exit the emulator");
					Console::WriteLine("go - trace mode");
					Console::WriteLine("stop - single step mode");
					Console::WriteLine(". - single step");
					Console::WriteLine("show - display opcodes/etc");
					Console::WriteLine("hide - don't show opcodes/etc");
					Console::WriteLine("ram - show RAM contents");
					Console::WriteLine("regs - list registers");
					Console::WriteLine("bps - list breakpoints");
					Console::WriteLine("board - display board");
					Console::WriteLine("coordinates: A B C D E F G H 1 2 3 4 5 6 7 8");
					Console::WriteLine("commands: TIME SET B/W - RANK CE RESET ENTER");
					Console::WriteLine("numbers: 0 9");

					doex=false;
				} else if(!String::Compare(line,"quit",true)) {
					doex=false;
					doit=false;
				} else if(!String::Compare(line,"int",true)) {		//simulate interrupt
					f8.irq_request=1;
				} else if(!String::Compare(line,"reset",true)) {		//simulate reset
					f8_reset(0);
				} else if(!String::Compare(line,"keys",true)) {
					Console::WriteLine("keycol={0}, keyrow={1}",keycol,keyrow);
				} else if(!String::Compare(line,".")) {
					singlestep=true;
					doex=true;
				} else if(!String::Compare(line,"j1")) {//jumps for testing
					f8.pc0=0x417;
				} else if(!String::Compare(line,"j2")) {
					f8.pc0=0x5e6;
				} else if(!String::Compare(line,"go",true)) {
					singlestep=false;
					doex=true;
				} else if(!String::Compare(line,"stop",true)) {
					singlestep=true;
					doex=true;
				} else if(!String::Compare(line,"show",true)) {
					show=true;
					doex=true;
				} else if(!String::Compare(line,"hide",true)) {
					show=false;
					doex=true;
				} else if(!String::Compare(line,"bps",true)) {
					doex=false;
					for(int i=0;i<NUMBPS;i++)
						Console::WriteLine(bps[i].ToString("X2"));
				} else if(!String::Compare(line,"ram",true)) {
					doex=false;
					for(int i=0;i<RAMSIZE;i++) {
						if(i%16==0)
							Console::WriteLine("");
						Console::Write(String::Concat(ram[i].ToString("X2")," "));
					}
					Console::WriteLine("");
				} else if(!String::Compare(line,"board",true)) {
					doex=false;
					for(int j=128;j>0;j-=16) {
						for(int i=0;i<8;i++)
							Console::Write(piece(ram[j+i-16]));
						Console::WriteLine("");
					}
				} else if(!String::Compare(line,"regs",true)) {
					doex=false;
					Console::WriteLine(String::Concat("PC0:$",f8.pc0.ToString("X4")," PC1:$",f8.pc1.ToString("X4")));
					Console::WriteLine(String::Concat("DC0:$",f8.dc0.ToString("X4")," DC1:$",f8.dc1.ToString("X4")));
					Console::WriteLine(String::Concat("Acc:$",f8.a.ToString("X4"),  "   W:$",f8.w.ToString("X4")));
					Console::WriteLine(String::Concat(" IS:$",f8.is.ToString("X4")));
					for(int i=0;i<64;i++) {
						Console::Write(String::Concat("R",i.ToString("D2"),":$",f8.r[i].ToString("X2"),"  "));
						if(((i+1)%8)==0)
							Console::WriteLine("");
					}
					Console::WriteLine("");
				} else if(!String::Compare(line,"timer",true)) {
					doex=false;
					Console::WriteLine("running={0}, int enabled={1}, divisor={2}, modulon={3}, irq={4}, count={5}, subcount={6}",f8.timerrunning,f8.timerintenabled,f8.timerdivisor,f8.timermodulon,f8.timerirq,f8.timercount,f8.timersubcount);
					Console::WriteLine("");
				} else if(!String::Compare(line," ",true)) {	//test command to clear the current key
					keyrow=9;
					keycol=0x0;
					singlestep=false;

				} else if(!String::Compare(line,"A",true)) {
					keyrow=1;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"1")) {
					keyrow=1;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"B",true)) {
					keyrow=1;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"2")) {
					keyrow=1;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"C",true)) {
					keyrow=1;
					keycol=0x40;
					singlestep=false;
				} else if(!String::Compare(line,"3")) {
					keyrow=1;
					keycol=0x40;
					singlestep=false;
				} else if(!String::Compare(line,"D",true)) {
					keyrow=2;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"4")) {
					keyrow=2;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"E",true)) {
					keyrow=2;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"5")) {
					keyrow=2;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"F",true)) {
					keyrow=2;
					keycol=0x40;
					singlestep=false;
				} else if(!String::Compare(line,"6")) {
					keyrow=2;
					keycol=0x40;
					singlestep=false;
				} else if(!String::Compare(line,"G",true)) {
					keyrow=3;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"7")) {
					keyrow=3;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"H",true)) {
					keyrow=3;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"8")) {
					keyrow=3;
					keycol=0x20;
					singlestep=false;

				} else if(!String::Compare(line,"CE",true)) {
					keyrow=3;
					keycol=0x80;
					singlestep=false;
				} else if(!String::Compare(line,"TIME",true)) {
					keyrow=2;
					keycol=0x80;
					singlestep=false;
				} else if(!String::Compare(line,"RANK",true)) {
					keyrow=1;
					keycol=0x80;
					singlestep=false;
				} else if(!String::Compare(line,"ENTER",true)) {
					keyrow=0;
					keycol=0x80;
					singlestep=false;
				} else if(!String::Compare(line,"-",true)) {
					keyrow=0;
					keycol=0x20;
					singlestep=false;
				} else if(!String::Compare(line,"B/W",true)) {
					keyrow=0;
					keycol=0x40;
					singlestep=false;

				} else if(!String::Compare(line,"0",true)) {
					keyrow=0;
					keycol=0x10;
					singlestep=false;
				} else if(!String::Compare(line,"9",true)) {
					keyrow=3;
					keycol=0x40;
					singlestep=false;
				} else if(!String::Compare(line,"SET",true)) {
					keyrow=3;
					keycol=0x40;
					singlestep=false;
				} else {
					Console::WriteLine("unknown command");
					doex=false;
				}
			}
			line=nullptr;
			if(doex)
				f8_execute(sw);

		}
//	} catch(Exception^) {
//		int bp=0;
//	}

#if DEBUGFILE
	sw->Close();//close debug file
#endif

#if DEBUGPC0
	psw->Close();
#endif


	fi = gcnew FileInfo( "f8coverage.txt" );
	fi->Delete();

	//dump the coverage map to file
	fs=gcnew FileStream("f8coverage.txt",FileMode::OpenOrCreate);
	sw=gcnew StreamWriter(fs);
	sw->WriteLine("1=read CPU opcode");
	sw->WriteLine("2=read CPU operand");
	sw->WriteLine("4=read data byte");
	sw->WriteLine("8=wrote data byte");
	sw->WriteLine("");
	sw->WriteLine("addr usage count");
	for(int i=0;i<ROMSIZE;i++) {
		sw->WriteLine("{0:X4}   {1}   {2}",i,access[i],numaccesses[i]);
//		sw->WriteLine(String::Concat(i.ToString("X4")," ",access[i].ToString("X2"),"    "));
	}
	sw->Close();


	fi = gcnew FileInfo( "f8ram.txt" );
	fi->Delete();

	//dump RAM to file
	fs=gcnew FileStream("f8ram.txt",FileMode::OpenOrCreate);
	sw=gcnew StreamWriter(fs);
	for(int i=0;i<RAMSIZE;i++) {
		sw->WriteLine(String::Concat(i.ToString("X4")," ",ram[i].ToString("X2"),"    "));
	}
	sw->Close();

	Console::WriteLine("done");
    return 0;
}
